# -*- coding: utf-8 -*-

# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import datetime
import os
import tempfile

from oslotest import base

import os_service_types.service_types


class TestCase(base.BaseTestCase):
    """Base test case class before singleton protection is added."""

    def setUp(self):
        super(TestCase, self).setUp()

        self.builtin_content = os_service_types.service_types.BUILTIN_DATA
        self.builtin_version = self.builtin_content['version']

        # Set up copies of the data so that we can verify that we got the
        # copy of it we think we should.
        self.remote_version = datetime.datetime.utcnow().isoformat()
        self.remote_content = copy.deepcopy(self.builtin_content)
        self.remote_content['version'] = self.remote_version


class ServiceDataMixin(object):

    scenarios = [
        ('compute', dict(
            service_type='compute', official='compute', aliases=[],
            all_types=['compute'],
            api_reference='compute', api_reference_project=None,
            is_secondary=False, all_services=['compute'],
            is_known=True, is_alias=False, is_official=True, project='nova')),
        ('volumev2', dict(
            service_type='volumev2', official='block-storage', aliases=[],
            all_types=['block-storage', 'volumev3', 'volumev2', 'volume',
                       'block-store'],
            api_reference='block-storage', api_reference_project=None,
            is_known=True, is_alias=True, is_official=False,
            is_secondary=False, all_services=['block-storage'],
            project='cinder')),
        ('volumev3', dict(
            service_type='volumev3', official='block-storage', aliases=[],
            all_types=['block-storage', 'volumev3', 'volumev2', 'volume',
                       'block-store'],
            api_reference='block-storage', api_reference_project=None,
            is_known=True, is_alias=True, is_official=False,
            is_secondary=False, all_services=['block-storage'],
            project='cinder')),
        ('block-storage', dict(
            service_type='block-storage', official='block-storage',
            all_types=['block-storage', 'volumev3', 'volumev2', 'volume',
                       'block-store'],
            api_reference='block-storage', api_reference_project=None,
            aliases=['volumev3', 'volumev2', 'volume', 'block-store'],
            is_known=True, is_alias=False, is_official=True,
            is_secondary=False, all_services=['block-storage'],
            project='cinder')),
        ('block_storage', dict(
            service_type='block_storage', official='block-storage',
            all_types=['block-storage', 'volumev3', 'volumev2', 'volume',
                       'block-store'],
            api_reference='block-storage', api_reference_project=None,
            aliases=['volumev3', 'volumev2', 'volume', 'block-store'],
            is_known=True, is_alias=False, is_official=True,
            is_secondary=False, all_services=['block-storage'],
            project='cinder')),
        ('network', dict(
            service_type='network', official='network', aliases=[],
            all_types=['network'],
            api_reference='network', api_reference_project='neutron-lib',
            is_known=True, is_alias=False, is_official=True,
            is_secondary=False, all_services=['network'],
            project='neutron')),
        ('placement', dict(
            service_type='placement', official='placement', aliases=[],
            all_types=['placement'], all_services=['placement'],
            api_reference='placement', api_reference_project=None,
            is_known=True, is_alias=False, is_official=True,
            is_secondary=False, project='placement')),
        ('missing', dict(
            service_type='missing', official=None,
            aliases=[], all_services=[],
            all_types=['missing'],
            api_reference=None, api_reference_project=None,
            is_known=False, is_alias=False, is_official=False,
            is_secondary=False,
            project=None)),
    ]

    def test_get_service_type(self):
        if self.official:
            self.assertEqual(
                self.official,
                self.service_types.get_service_type(self.service_type))
        else:
            self.assertIsNone(
                self.service_types.get_service_type(self.service_type))

    def test_get_service_type_permissive(self):
        self.assertEqual(
            self.official or self.service_type,
            self.service_types.get_service_type(
                self.service_type, permissive=True))

    def test_get_aliases(self):
        self.assertEqual(
            self.aliases,
            self.service_types.get_aliases(self.service_type))

    def test_is_known(self):
        self.assertEqual(
            self.is_known,
            self.service_types.is_known(self.service_type))

    def test_is_alias(self):
        self.assertEqual(
            self.is_alias,
            self.service_types.is_alias(self.service_type))

    def test_is_official(self):
        self.assertEqual(
            self.is_official,
            self.service_types.is_official(self.service_type))

    def test_get_project_name(self):
        self.assertEqual(
            self.project,
            self.service_types.get_project_name(self.service_type))

    def test_get_service_data(self):
        service_data = self.service_types.get_service_data(self.service_type)
        # TODO(mordred) Once all the docs have been aligned, remove
        # self.api_reference and replace with self.service_type
        api_url = 'https://developer.openstack.org/api-ref/{api_reference}/'

        # Tests self.official here, since we expect to get data back for all
        # official projects, regardless of service_type being an alias or not
        if not self.official:
            self.assertIsNone(service_data)
        else:
            self.assertIsNotNone(service_data)
            self.assertEqual(self.project, service_data['project'])
            self.assertEqual(self.official, service_data['service_type'])
            self.assertEqual(
                api_url.format(api_reference=self.api_reference),
                service_data['api_reference'])

    def test_get_official_service_data(self):
        service_data = self.service_types.get_official_service_data(
            self.service_type)
        # TODO(mordred) Once all the docs have been aligned, remove
        # self.api_reference and replace with self.service_type
        api_url = 'https://developer.openstack.org/api-ref/{api_reference}/'

        # Tests self.is_official here, since we expect only get data back for
        # official projects.
        if not self.is_official:
            self.assertIsNone(service_data)
        else:
            self.assertIsNotNone(service_data)
            self.assertEqual(self.project, service_data['project'])
            self.assertEqual(self.official, service_data['service_type'])
            self.assertEqual(
                api_url.format(api_reference=self.api_reference),
                service_data['api_reference'])

    def test_empty_project_error(self):
        if not self.project:
            self.assertRaises(
                ValueError,
                self.service_types.get_service_data_for_project,
                self.project)

    def test_get_service_data_for_project(self):
        if self.is_secondary:
            self.skipTest("Secondary services have no project mapping")
            return
        elif not self.project:
            self.skipTest("Empty project is invalid but tested elsewhere.")
            return

        service_data = self.service_types.get_service_data_for_project(
            self.project)
        # TODO(mordred) Once all the docs have been aligned, remove
        # self.api_reference and replace with self.service_type
        api_url = 'https://developer.openstack.org/api-ref/{api_reference}/'

        self.assertIsNotNone(service_data)
        if self.api_reference_project:
            self.assertEqual(
                self.api_reference_project,
                service_data['api_reference_project'])
        else:
            self.assertEqual(self.project, service_data['project'])
        self.assertEqual(self.official, service_data['service_type'])
        self.assertEqual(
            api_url.format(api_reference=self.api_reference),
            service_data['api_reference'])

    def test_get_all_types(self):
        self.assertEqual(
            self.all_types,
            self.service_types.get_all_types(self.service_type))

    def test_all_get_service_data_for_project(self):
        if not self.project:
            self.skipTest("Empty project is invalid but tested elsewhere.")
            return

        all_data = self.service_types.get_all_service_data_for_project(
            self.project)
        for (index, data) in enumerate(all_data):
            self.assertEqual(
                data,
                self.service_types.get_service_data(self.all_services[index]))


class TemporaryFileMixin(base.BaseTestCase):

    def create_temp_file(self, mode='w', suffix='', prefix='tmp', dir=None,
                         text=False, delete=True):
        fd, name = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir,
                                    text=text)
        fd = os.fdopen(fd, mode)
        if delete:
            self.addCleanup(self._delete_temp, fd, name)
        return fd, name

    def _delete_temp(self, fd, name):
        fd.close()
        os.unlink(name)



# Copyright 2017 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import os_service_types.data
from os_service_types import exc

__all__ = ['ServiceTypes']

BUILTIN_DATA = os_service_types.data.read_data('service-types.json')
SERVICE_TYPES_URL = "https://service-types.openstack.org/service-types.json"


def _normalize_type(service_type):
    if service_type:
        return service_type.replace('_', '-')


class ServiceTypes(object):
    """Encapsulation of the OpenStack Service Types Authority data.

    The Service Types Authority data will be either pulled from its remote
    location or from local files as is appropriate.

    If the user passes a Session, remote data will be fetched. If the user
    does not do that, local builtin data will be used.

    :param session: An object that behaves like a `requests.sessions.Session`
        or a `keystoneauth1.session.Session` that provides a get method
        and returns an object that behaves like a `requests.models.Response`.
        Optional. If session is omitted, no remote actions will be performed.
    :param bool only_remote: By default if there is a problem fetching data
        remotely the builtin data will be returned as a fallback. only_remote
        will cause remote failures to raise an error instead of falling back.
        Optional, defaults to False.
    :param bool warn: Emit warnings when a non-official service_type is
        provided. This provides an easy way for consuming applications to
        warn users when they are using old types.
    :raises ValueError: If session is None and only_remote is True
    :raises IOError: If session is given and only_remote is True and there is
        an error fetching remote data.
    """

    def __init__(self, session=None, only_remote=False, warn=False):
        if not session and only_remote:
            raise ValueError(
                "only_remote was requested but no Session was provided.")
        self._service_types_data = BUILTIN_DATA
        self._warn = warn
        if session:
            try:
                response = session.get(SERVICE_TYPES_URL)
                response.raise_for_status()
                self._service_types_data = response.json()
            except IOError:
                # If we can't fetch, fall backto BUILTIN
                if only_remote:
                    raise

    def _canonical_project_name(self, name):
        "Convert repo name to project name."
        if name is None:
            raise ValueError("Empty project name is not allowed")
        # Handle openstack/ prefix going away from STA data
        return name.rpartition('/')[-1]

    @property
    def url(self):
        "The URL from which the data was retrieved."
        return SERVICE_TYPES_URL

    @property
    def version(self):
        "The version of the data."
        return self._service_types_data['version']

    @property
    def forward(self):
        "Mapping service-type names to their aliases."
        return copy.deepcopy(self._service_types_data['forward'])

    @property
    def reverse(self):
        "Mapping aliases to their service-type names."
        return copy.deepcopy(self._service_types_data['reverse'])

    @property
    def services(self):
        "Full service-type data listing."
        return copy.deepcopy(self._service_types_data['services'])

    @property
    def all_types_by_service_type(self):
        "Mapping of official service type to official type and aliases."
        return copy.deepcopy(
            self._service_types_data['all_types_by_service_type'])

    @property
    def primary_service_by_project(self):
        "Mapping of project name to the primary associated service."
        return copy.deepcopy(
            self._service_types_data['primary_service_by_project'])

    @property
    def service_types_by_project(self):
        "Mapping of project name to a list of all associated service-types."
        return copy.deepcopy(
            self._service_types_data['service_types_by_project'])

    def get_official_service_data(self, service_type):
        """Get the service data for an official service_type.

        :param str service_type: The official service-type to get data for.
        :returns dict: Service data for the service or None if not found.
        """
        service_type = _normalize_type(service_type)
        for service in self._service_types_data['services']:
            if service_type == service['service_type']:
                return service
        return None

    def get_service_data(self, service_type):
        """Get the service data for a given service_type.

        :param str service_type: The service-type or alias to get data for.
        :returns dict: Service data for the service or None if not found.
        """
        service_type = self.get_service_type(service_type)
        if not service_type:
            return None
        return self.get_official_service_data(service_type)

    def is_official(self, service_type):
        """Is the given service-type an official service-type?

        :param str service_type: The service-type to test.
        :returns bool: True if it's an official type, False otherwise.
        """
        return self.get_official_service_data(service_type) is not None

    def is_alias(self, service_type):
        """Is the given service-type an alias?

        :param str service_type: The service-type to test.
        :returns bool: True if it's an alias type, False otherwise.
        """
        service_type = _normalize_type(service_type)
        return service_type in self._service_types_data['reverse']

    def is_known(self, service_type):
        """Is the given service-type an official type or an alias?

        :param str service_type: The service-type to test.
        :returns bool: True if it's a known type, False otherwise.
        """
        return self.is_official(service_type) or self.is_alias(service_type)

    def is_match(self, requested, found):
        """Does a requested service-type match one found in the catalog?

        A requested service-type matches a service-type in the catalog if
        it is either a direct match, if the service-type in the catalog is
        an official type and the requested type is one of its aliases, or
        if the requested type is an official type and the type in the catalog
        is one of its aliases.

        A requested alias cannot match a different alias because there are
        historical implications related to versioning to some historical
        aliases that cannot be safely reasoned about in an automatic fashion.

        :param str requested: A service-type that someone is looking for.
        :param str found: A service-type found in a catalog

        :returns bool: True if the service-type being requested matches the
            entry in the catalog. False if it does not.
        """
        # Exact match
        if requested == found:
            return True

        # Found is official type, requested is one of its aliases
        if requested in self.get_aliases(found):
            return True

        # Found is an alias, requested is an official type
        if requested == self.get_service_type(found):
            return True

        return False

    def get_aliases(self, service_type):
        """Returns the list of aliases for a given official service-type.

        :param str service_type: An official service-type.
        :returns list: List of aliases, or empty list if there are none.
        """
        service_type = _normalize_type(service_type)
        return self._service_types_data['forward'].get(service_type, [])

    def get_service_type(self, service_type, permissive=False):
        """Given a possible service_type, return the official type.

        :param str service_type: A potential service-type.
        :param bool permissive:
            Return the original type if the given service_type is not found.
        :returns str: The official service-type, or None if there is no match.
        """
        service_type = _normalize_type(service_type)
        if self.is_official(service_type):
            return service_type
        official = self._service_types_data['reverse'].get(service_type)
        if permissive and official is None:
            if self._warn:
                exc.warn(
                    exc.UnofficialUsageWarning,
                    given=service_type)
            return service_type
        if self._warn:
            exc.warn(
                exc.AliasUsageWarning, given=service_type, official=official)
        return official

    def get_all_types(self, service_type):
        """Get a list of official types and all known aliases.

        :param str service_type: The service-type or alias to get data for.
        :returns dict: Service data for the service or None if not found.
        """
        service_type = _normalize_type(service_type)
        if not self.is_known(service_type):
            return [service_type]
        return self.all_types_by_service_type[
            self.get_service_type(service_type)]

    def get_project_name(self, service_type):
        """Return the OpenStack project name for a given service_type.

        :param str service_type: An official service-type or alias.
        :returns str: The OpenStack project name or None if there is no match.
        """
        service_type = _normalize_type(service_type)
        service = self.get_service_data(service_type)
        if service:
            return service['project']
        return None

    def get_service_data_for_project(self, project_name):
        """Return the service information associated with a project.

        :param name: A repository or project name in the form
            ``'openstack/{project}'`` or just ``'{project}'``.
        :type name: str
        :raises ValueError: If project_name is None
        :returns: dict or None if not found
        """
        project_name = self._canonical_project_name(project_name)
        return self.primary_service_by_project.get(project_name)

    def get_all_service_data_for_project(self, project_name):
        """Return the information for every service associated with a project.

        :param name: A repository or project name in the form
            ``'openstack/{project}'`` or just ``'{project}'``.
        :type name: str
        :raises ValueError: If project_name is None
        :returns: list of dicts
        """
        data = []
        for service_type in self.service_types_by_project.get(
                project_name, []):
            data.append(self.get_service_data(service_type))
        return data



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_builtin
------------

Tests for `ServiceTypes` class builtin data.

oslotest sets up a TempHomeDir for us, so there should be no ~/.cache files
available in these tests.
"""
from testscenarios import load_tests_apply_scenarios as load_tests  # noqa

import os_service_types
from os_service_types.tests import base


class TestBuiltin(base.TestCase, base.ServiceDataMixin):

    def setUp(self):
        super(TestBuiltin, self).setUp()
        # Make an object with no network access
        self.service_types = os_service_types.ServiceTypes()

    def test_builtin_version(self):
        self.assertEqual(self.builtin_version, self.service_types.version)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_data
------------

Tests for `os_service_types.data` module.

"""

import json

import six

from os_service_types import data
from os_service_types.tests import base


if six.PY2:
    # Python 2 has not FileNotFoundError exception
    FileNotFoundError = IOError


class TestData(base.TestCase, base.TemporaryFileMixin):

    def setUp(self):
        super(TestData, self).setUp()

    def test_load(self):
        json_data = {'some_key': 'some_value'}
        filename = self.create_json(json_data)
        actual_data = data.read_data(filename)
        self.assertEqual(json_data, actual_data)

    def test_load_service_types(self):
        json_data = data.read_data('service-types.json')
        for key in ["all_types_by_service_type", "forward",
                    "primary_service_by_project", "reverse"]:
            self.assertIn(key, json_data)

    def test_load_non_existing(self):
        self.assertRaises(FileNotFoundError, data.read_data,
                          '/non-existing-file')

    def create_json(self, json_data):
        fd, name = self.create_temp_file(suffix='.json')
        with fd:
            json.dump(json_data, fd)
        return name



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_warn
---------

Tests for warnings
"""
import warnings

import os_service_types
from os_service_types import exc
from os_service_types.tests import base


class TestWarnOn(base.TestCase):

    def setUp(self):
        super(TestWarnOn, self).setUp()
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        self.service_types = os_service_types.ServiceTypes(warn=True)

    def test_warning_emitted_on_alias(self):
        with warnings.catch_warnings(record=True) as w:
            self.service_types.get_service_type('volumev2')
            self.assertEqual(1, len(w))
            self.assertTrue(issubclass(w[-1].category, exc.AliasUsageWarning))

    def test_warning_not_emitted_on_official(self):
        with warnings.catch_warnings(record=True) as w:
            self.service_types.get_service_type('block-storage')
            self.assertEqual(0, len(w))


class TestWarnOff(base.TestCase):

    def setUp(self):
        super(TestWarnOff, self).setUp()
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        self.service_types = os_service_types.ServiceTypes()

    def test_warning_not_emitted_on_alias(self):
        with warnings.catch_warnings(record=True) as w:
            self.service_types.get_service_type('volumev2')
            self.assertEqual(0, len(w))

    def test_warning_not_emitted_on_official(self):
        with warnings.catch_warnings(record=True) as w:
            self.service_types.get_service_type('block-storage')
            self.assertEqual(0, len(w))



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_match
----------

Tests for is_match logic

oslotest sets up a TempHomeDir for us, so there should be no ~/.cache files
available in these tests.
"""
from testscenarios import load_tests_apply_scenarios as load_tests  # noqa

import os_service_types
from os_service_types.tests import base


class TestMatch(base.TestCase):

    scenarios = [
        ('match-official', dict(
            requested='compute', found='compute', is_match=True)),
        ('direct-match-unknown', dict(
            requested='unknown', found='unknown', is_match=True)),
        ('volumev2-finds-block', dict(
            requested='volumev2', found='block-storage', is_match=True)),
        ('volumev3-finds-block', dict(
            requested='volumev3', found='block-storage', is_match=True)),
        ('block-finds-volumev2', dict(
            requested='block-storage', found='volumev2', is_match=True)),
        ('block-finds-volumev3', dict(
            requested='block-storage', found='volumev3', is_match=True)),
        ('volumev2-not-volumev3', dict(
            requested='volumev2', found='volumev3', is_match=False)),
        ('non-match', dict(
            requested='unknown', found='compute', is_match=False)),
    ]

    def setUp(self):
        super(TestMatch, self).setUp()
        # Make an object with no network access
        self.service_types = os_service_types.ServiceTypes()

    def test_is_match(self):
        self.assertEqual(
            self.is_match,
            self.service_types.is_match(self.requested, self.found))
